dataset_name = "Yeast"
num_classes = 2
num_features = 46
model = "cgc_classification"
evaluation = "classification"
optimizer = "adams"  # defaults to adams optimizer
net_dimension = 364
learning_rate = 0.0005
dataset_save_path = "t./data"
test_ratio = 20
batch_size = 64
shuffle = True
gamma = 51
num_epoch = 300
num_cv = 5
result_save_dir = "./Yeast"
data_loader = "StandardTUD"
model_save_freq = 1
